@@ -8,6 +8,7 @@ import math
 import os
 import sys
 import time
+import threading
 from warnings import warn
 
 from PythonReports import barcode, drivers
@@ -392,7 +393,7 @@ class ReportElement(Structure):
     # section: containing section object
     # template: template element
     # style: dictionary of style attributes
-    # printable: True (non-printable elements cannot be processed at present)
+    # printable: False if the element is suppressed (by style printwhen)
     # tbox: box defined in the template
     # bbox: box with absolute sizes (computed from tbox and section sizes)
     # obox: output box, with absolute position and size values
@@ -439,6 +440,7 @@ class Section(list):
             (prt.Rectangle, prp.Rectangle),
             (prt.Image, prp.Image),
             (prt.BarCode, prp.BarCode),
+            (prt.Outline, prp.Outline),
         ))
 
     # Bar Code drivers
@@ -448,6 +450,10 @@ class Section(list):
         "2of5i": barcode.code2of5i,
     }
 
+    # Tag names for printable output elements (with dimension boxes)
+    PRINTABLE_ELEMENT_TAGS = frozenset(("field",
+        "line", "rectangle", "image", "barcode"))
+
     # printability is set by .build
     printable = None
 
@@ -487,7 +493,8 @@ class Section(list):
         self.template = template
         self.box = self.tbox = Box.from_element(template.find("box"))
         self.resizeable = False
-        self.subreports_before = self.subreports_after = () # filled by .build
+        # subreports and bookmarks lists are created in .build()
+        self.subreports_before = self.subreports_after = self.bookmarks = ()
         self.template2element = {}
         self.vertical_segment_layout = self.has_floating_boxes = None
         self.create_vertical_segment_layout(template)
@@ -501,7 +508,8 @@ class Section(list):
 
         """
         _elements = [(_item, Box.from_element(_item.find("box")))
-            for _item in template if self.is_body_element(_item)]
+            for _item in template
+            if _item.tag in self.PRINTABLE_ELEMENT_TAGS]
 
         self.has_floating_boxes = any(_it[1].float and
             (_it[1].y >= 0) and (_it[1].height >= 0) for _it in _elements)
@@ -736,13 +744,19 @@ class Section(list):
         _default_element_style = [self.compose_style(context,
             ("font", "color", "bgcolor"), self.iter_styles())]
         _subreports = []
+        _bookmarks = []
         for _item in _elements:
             if _item.tag == "subreport":
                 _when = _item.get("when")
                 if (not _when) or context.eval(_when, _item):
                     _subreports.append((_item.get("seq"), _item))
                 continue
-            if not self.is_body_element(_item):
+            if _item.tag == "outline":
+                _bookmarks.append(Structure(template=_item,
+                    name=self.builder.generate_id(),
+                    title=context.eval(_item.get("title"), _item)))
+                continue
+            if _item.tag not in self.PRINTABLE_ELEMENT_TAGS:
                 continue
             _element = ReportElement(section=self, template=_item)
             self.template2element[_item] = _element
@@ -773,16 +787,7 @@ class Section(list):
             if _item[0] < 0)
         self.subreports_after = tuple(_item[1] for _item in _subreports
             if _item[0] > 0)
-
-    @staticmethod
-    def is_body_element(item):
-        """@return: whether item represents body element or not.
-
-        @param item: template section element.
-
-        """
-        # TODO convert it to element's or template's property or method
-        return item.tag in ("field", "line", "rectangle", "image", "barcode")
+        self.bookmarks = tuple(_bookmarks)
 
     def build_barcode(self, element):
         """Compute sripe widths and box size/position for barcode element
@@ -804,10 +809,9 @@ class Section(list):
         if element.template.get("vertical"):
             (_min_height, _min_width) = (_min_width, _min_height)
         _bbox = element.tbox.copy()
-        if _bbox.width < _min_width:
-            _bbox.width = _min_width
-        if _bbox.height < _min_height:
-            _bbox.height = _min_height
+        # Always apply minimums here, actual sizes calculated in .refill()
+        _bbox.width = _min_width
+        _bbox.height = _min_height
         _bbox.place_x(self.box)
         _bbox.place_y(self.box)
         element.bbox = _bbox
@@ -885,8 +889,18 @@ class Section(list):
         if self.resizeable:
             _height = self.tbox.height
             for _element in self:
+                if _element.template.tag == "barcode":
+                    # For barcodes, bbox is minimum allowed.
+                    # If there is bigger fixed size in the template,
+                    # use that size instead of bbox minimum.
+                    _bbox = _element.bbox.copy()
+                    if _element.tbox.width > _bbox.width:
+                        _bbox.width = _element.tbox.width
+                    if _element.tbox.height > _bbox.height:
+                        _bbox.height = _element.tbox.height
+                else:
+                    _bbox = _element.bbox
                 # Note: bbox vertical dimensions are relative yet
-                _bbox = _element.bbox
                 if (_bbox.height < 0) and (_element.template.tag == "image") \
                 and (_element.template.get("scale") == "grow"):
                     # box height is relative to section size
@@ -902,7 +916,6 @@ class Section(list):
                 elif _bbox.height >= 0:
                     _elem_height = _bbox.y + _bbox.height
                 else:
-                    # FIXME is it for _bbox.height < 0?
                     # fixed space from top and bottom, unknown size
                     _elem_height = _bbox.y + 1 - _bbox.height
                 if _elem_height > _height:
@@ -967,6 +980,28 @@ class Section(list):
             if _element.template.tag == "barcode":
                 # update encoded symbol and bbox from current text if needed
                 self.build_barcode(_element)
+                # Do the final placement: align width, grow height
+                _pbox = _element.tbox.copy()
+                _pbox.place_x(self.box)
+                _pbox.place_y(self.box)
+                _obox = _element.bbox.copy() # Minimal allowed dimensions
+                if not _element.template.get("grow"):
+                    # Align both dimensions.
+                    _obox.align_x(_pbox)
+                    _obox.align_y(_pbox)
+                elif _element.template.get("vertical"):
+                    # Align vertically, grow horizontally.
+                    _obox.align_y(_pbox)
+                    if _obox.width < _pbox.width:
+                        _obox.width = _pbox.width
+                else:
+                    # Align horizontally, grow vertically.
+                    _obox.align_x(_pbox)
+                    if _obox.height < _pbox.height:
+                        _obox.height = _pbox.height
+                _element.obox = _obox
+                # No more adjustments for barcodes.
+                continue
             _bbox = _element.bbox
             if _sbox:
                 # vertical position changed.  recalc from template.
@@ -998,12 +1033,22 @@ class Section(list):
             page: printout.Page object
 
         """
+        for _element in self.bookmarks:
+            _template = _element.template
+            (_prp_type, _prp_attrs) = self.PRINTOUTS[_template.tag]
+            _attrib = dict([(_name, _template.get(_name, ""))
+                for _name in _prp_attrs])
+            # 01-apr-2017 The only bookmarks type is Outline
+            _attrib.update(dict(x=self.box.x, y=self.box.y,
+                name=_element.name, title=_element.title,
+            ))
+            _prp_element = SubElement(page, _prp_type.tag, _attrib)
         for _element in self:
             # build definition for printout element
             _template = _element.template
             (_prp_type, _prp_attrs) = self.PRINTOUTS[_template.tag]
             _prp_tag = _prp_type.tag
-            _attrib = dict([(_name, _template.get(_name))
+            _attrib = dict([(_name, _template.get(_name, ""))
                 for _name in _prp_attrs])
             # add style attributes (must be done before constructor is called)
             if _prp_tag == "text":
@@ -1182,6 +1227,9 @@ class Builder(object):
 
     """
 
+    # An ID generator state
+    gen_id = 0
+    gen_id_lock = threading.Lock()
     ### following properties hold builder state.  reinitialized by each .run()
     # current page position
     cur_y = 0
@@ -1281,6 +1329,17 @@ class Builder(object):
         return "<%s@%x:%r>" % (self.__class__.__name__, id(self),
             os.path.basename(self.template.filename))
 
+    @classmethod
+    def generate_id(cls):
+        """Return a new unique element ID string"""
+        cls.gen_id_lock.acquire()
+        try:
+            cls.gen_id += 1
+            _rv = "i%08X" % cls.gen_id
+        finally:
+            cls.gen_id_lock.release()
+        return _rv
+
     def filepath(self, *path):
         """Return normalized absolute pathname
 
@@ -1356,7 +1415,7 @@ class Builder(object):
         while _descend:
             _next_level = []
             for _item in _descend:
-                for _child in getchildren(_item):
+                for _child in self.template.getchildren(_item):
                     if _child.tag in ("group", "columns"):
                         _next_level.append(_child)
                     if _child.tag in (
@@ -1934,6 +1993,104 @@ class Builder(object):
             _frame.top = ypos
             _frame = _frame.child
 
+    def build_embedded_template(self, element):
+        """Return a template ElementTree for embedded subreport
+
+        Embedded subreports use imports, fonts, named data blocks,
+        and page layout parameters from the main template.
+
+        Build a complete Template ElementTree for a subreport
+        from a combination of own template parts and the element contents.
+
+        """
+        _this_root = self.template.getroot()
+        _new_root = self.template.copy(_this_root)
+        # Make a shallow copy of all top-level elements.
+        for _elem in self.template.getchildren(_this_root):
+            if _elem.tag in ("import", "font", "data", "layout"):
+                _new_root.append(self.template.copy(_elem))
+        _new_layout = _new_root.find("layout")
+        for _elem in self.template.getchildren(element):
+            if _elem.tag in ("parameter", "variable"):
+                _new_root.insert(0, self.template.copy(_elem))
+            else:
+                _new_layout.append(_elem)
+        _rv = ElementTree(self.template.root_validator, _new_root)
+        _rv.filename = self.template.filename
+        _rv.validate()
+        return _rv
+
+    def get_subreport_builder(self, element):
+        """Return a Builder object for a subreport element
+
+        Parameters:
+            element: a "subreport" element from the report template
+
+        Return: tuple (is_inline, builder)
+
+        """
+        _embedded = element.get("embedded")
+        _is_inline = bool(_embedded or element.get("inline"))
+        if element in self.subreports:
+            return (_is_inline, self.subreports[element])
+        _this_layout = self.template.find("layout")
+        if _embedded:
+            _prt = self.build_embedded_template(
+                self.template.embedded[_embedded])
+            _layout = _prt.find("layout")
+        else:
+            _prt_name = element.get("template")
+            _prt = prt.load(self.filepath(_prt_name))
+            if _is_inline:
+                # inlined report must have same page dimensions as this report
+                _pgsize = self.get_page_dimensions(self.template)
+                if self.get_page_dimensions(_prt) != _pgsize:
+                    raise XmlValidationError(
+                        "Page size does not match for inlined report \"%s\""
+                        % _prt_name, element=element)
+                _layout = _prt.find("layout")
+                assert _layout is not None # _prt is verified
+                _MARGIN_ATTRS = ("leftmargin", "topmargin",
+                    "rightmargin", "bottommargin")
+                _margins = tuple(tuple(_section.get(_margin, 0)
+                    for _margin in _MARGIN_ATTRS)
+                    for _section in (_layout, _this_layout))
+                if _margins[0] != _margins[1]:
+                    warn(XmlValidationWarning("Overriding page margins"
+                        " for subreport %s: (%s) => (%s)" % ((_prt_name,)
+                        + tuple(", ".join(map(str, _margin))
+                            for _margin in _margins)),
+                        element=_layout))
+                    for (_name, _value) in zip(_MARGIN_ATTRS, _margins[1]):
+                        _layout.set(_name, _value)
+            else:
+                _builder = Builder(_prt)
+        if _is_inline:
+            # Insert page headers and footers from own template.
+            # Actual rendering will be done by responsible builders,
+            # but subreport builder will need to know sizes
+            # to shrink page contents frame appropriately.
+            _page_frame = self.section_frames[None]
+            _outer_sections = _this_layout.findall("footer")
+            if sys.version_info[:2] < (2, 7):
+                # _ElementInterface instance has no attribute 'extend'
+                for _section in _outer_sections:
+                    _layout.append(_section)
+            else:
+                _layout.extend(_outer_sections)
+            for _section in reversed(_this_layout.findall("header")):
+                _layout.insert(0, _section)
+                _outer_sections.append(_section)
+            # create subreport builder
+            _builder = Builder(_prt)
+            # copied sections must be built by outer builders
+            if _outer_sections:
+                _builder.section_builders = dict(
+                    (_section, self.section_builders.get(_section, self))
+                    for _section in _outer_sections)
+        self.subreports[element] = _builder
+        return (_is_inline, _builder)
+
     def run_subreport(self, element, eject_frame):
         """Execute a subreport element
 
@@ -1957,61 +2114,7 @@ class Builder(object):
         _data = _context.eval(element.get("data"), element)
         if len(_data) < 1:
             return
-        _inline = element.get("inline")
-        # fetch or make a builder
-        if element in self.subreports:
-            _builder = self.subreports[element]
-        else:
-            _prt_name = element.get("template")
-            _prt = prt.load(self.filepath(_prt_name))
-            if _inline:
-                # inlined report must have same page dimensions as this report
-                _pgsize = self.get_page_dimensions(self.template)
-                if self.get_page_dimensions(_prt) != _pgsize:
-                    raise XmlValidationError(
-                        "Page size does not match for inlined report \"%s\""
-                        % _prt_name, element=element)
-                _layout = _prt.find("layout")
-                assert _layout is not None # _prt is verified
-                _this_layout = self.template.find("layout")
-                _MARGIN_ATTRS = ("leftmargin", "topmargin",
-                    "rightmargin", "bottommargin")
-                _margins = tuple(tuple(_section.get(_margin, 0)
-                    for _margin in _MARGIN_ATTRS)
-                    for _section in (_layout, _this_layout))
-                if _margins[0] != _margins[1]:
-                    warn(XmlValidationWarning("Overriding page margins"
-                        " for subreport %s: (%s) => (%s)" % ((_prt_name,)
-                        + tuple(", ".join(map(str, _margin))
-                            for _margin in _margins)),
-                        element=_layout))
-                    for (_name, _value) in zip(_MARGIN_ATTRS, _margins[1]):
-                        _layout.set(_name, _value)
-                # Insert page headers and footers from own template.
-                # Actual rendering will be done by responsible builders,
-                # but subreport builder will need to know sizes
-                # to shrink page contents frame appropriately.
-                _page_frame = self.section_frames[None]
-                _outer_sections = _this_layout.findall("footer")
-                if sys.version_info[:2] < (2, 7):
-                    # _ElementInterface instance has no attribute 'extend'
-                    for _section in _outer_sections:
-                        _layout.append(_section)
-                else:
-                    _layout.extend(_outer_sections)
-                for _section in _this_layout.findall("header"):
-                    _layout.insert(0, _section)
-                    _outer_sections.append(_section)
-                # create subreport builder
-                _builder = Builder(_prt)
-                # copied sections must be built by outer builders
-                if _outer_sections:
-                    _builder.section_builders = dict(
-                        (_section, self.section_builders.get(_section, self))
-                        for _section in _outer_sections)
-            else:
-                _builder = Builder(_prt)
-            self.subreports[element] = _builder
+        (_inline, _builder) = self.get_subreport_builder(element)
         # collect subreport arguments
         # Note: this is done before any new section is built
         #       to make sure current build context is not changed and
@@ -2288,7 +2391,7 @@ class Builder(object):
                 _frame.x = _xpos
                 _frame.column = 0
         # print all headers
-        for _frame in reversed(_eject_frames):
+        for _frame in _eject_frames:
             self.add_section(self.build_section(_frame.header))
 
     def register_eval(self, element):
@@ -2416,6 +2519,8 @@ class Builder(object):
             _prt_page = SubElement(_root, prp.Page.tag, _page_attrs)
             for _section in _page:
                 _section.output(_prt_page)
-        return ElementTree(prp.Printout, _root)
+        _rv = ElementTree(prp.Printout, _root)
+        _rv.validate()
+        return _rv
 
 # vim: set et sts=4 sw=4 :
